from .. import ImageNormalizer


def test_transform_results(self):
    img_size = 217
    target_size = 224
    crafter = ImageNormalizer(target_size=target_size)
    img_array = self.create_random_img_array(img_size, img_size)
    crafted_doc = crafter.craft(img_array)
    assert crafted_doc["blob"].shape == (224, 224, 3)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict

import numpy as np

from .. import BaseExecutor


class BaseRanker(BaseExecutor):
    """The base class for a `Ranker`"""

    def score(self, *args, **kwargs):
        raise NotImplementedError


class Chunk2DocRanker(BaseRanker):
    """ A :class:`Chunk2DocRanker` translates the chunk-wise score (distance) to the doc-wise score.

    In the query-time, :class:`Chunk2DocRanker` is an almost-always required component.
    Because in the end we want to retrieve top-k documents of given query-document not top-k chunks of
    given query-chunks. The purpose of :class:`Chunk2DocRanker` is to aggregate the already existed top-k chunks
    into documents.

    The key function here is :func:`score`.

    .. seealso::
        :mod:`jina.drivers.handlers.score`

    """

    required_keys = {'text'}  #: a set of ``str``, key-values to extracted from the chunk-level protobuf message

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.col_doc_id = 0
        self.col_chunk_id = 1
        self.col_query_chunk_id = 2
        self.col_score = 3

    def score(self, match_idx: 'np.ndarray', query_chunk_meta: Dict, match_chunk_meta: Dict) -> 'np.ndarray':
        """Translate the chunk-level top-k results into doc-level top-k results. Some score functions may leverage the
        meta information of the query, hence the meta info of the query chunks and matched chunks are given
        as arguments.

        :param match_idx: a [N x 4] numpy ``ndarray``, column-wise:

                - ``match_idx[:, 0]``: ``doc_id`` of the matched chunks, integer
                - ``match_idx[:, 1]``: ``chunk_id`` of the matched chunks, integer
                - ``match_idx[:, 2]``: ``chunk_id`` of the query chunks, integer
                - ``match_idx[:, 3]``: distance/metric/score between the query and matched chunks, float
        :param query_chunk_meta: the meta information of the query chunks, where the key is query chunks' ``chunk_id``,
            the value is extracted by the ``required_keys``.
        :param match_chunk_meta: the meta information of the matched chunks, where the key is matched chunks'
            ``chunk_id``, the value is extracted by the ``required_keys``.
        :return: a [N x 2] numpy ``ndarray``, where the first column is the matched documents' ``doc_id`` (integer)
                the second column is the score/distance/metric between the matched doc and the query doc (float).
        """
        _groups = self.group_by_doc_id(match_idx)
        r = []
        for _g in _groups:
            _doc_id, _doc_score = self._get_score(_g, query_chunk_meta, match_chunk_meta)
            r.append((_doc_id, _doc_score))
        return self.sort_doc_by_score(r)

    def group_by_doc_id(self, match_idx):
        """
        Group the ``match_idx`` by ``doc_id``
        :return: an iterator over the groups
        """
        return self._group_by(match_idx, self.col_doc_id)

    @staticmethod
    def _group_by(match_idx, col):
        # sort by ``col``
        _sorted_m = match_idx[match_idx[:, col].argsort()]
        _, _doc_counts = np.unique(_sorted_m[:, col], return_counts=True)
        # group by ``col``
        return np.split(_sorted_m, np.cumsum(_doc_counts))[:-1]

    def _get_score(self, match_idx, query_chunk_meta, match_chunk_meta, *args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def sort_doc_by_score(r):
        """
        Sort a list of (``doc_id``, ``score``) tuples by the ``score``.
        :return: an `np.ndarray` in the shape of [N x 2], where `N` in the length of the input list.
        """
        r = np.array(r, dtype=np.float64)
        r = r[r[:, -1].argsort()[::-1]]
        return r

    def get_doc_id(self, match_with_same_doc_id):
        return match_with_same_doc_id[0, self.col_doc_id]



import numpy as np

from .. import AudioSlicer, SlidingWindowAudioSlicer


class TestClass:
    def test_segment_mono(self):
        n_frames = 100
        frame_length = 2048
        signal_orig = np.random.randn(frame_length * n_frames)

        crafter = AudioSlicer(frame_length)
        crafted_chunks = crafter.craft(signal_orig, 0)

        assert len(crafted_chunks) == n_frames

    def test_segment_stereo(self):
        n_frames = 100
        frame_length = 2048
        signal_orig = np.random.randn(2, frame_length * n_frames)

        crafter = AudioSlicer(frame_length)
        crafted_chunks = crafter.craft(signal_orig, 0)

        assert len(crafted_chunks) == n_frames * 2

    def test_sliding_window_mono(self):
        n_frames = 100
        frame_length = 2048
        signal_orig = np.random.randn(frame_length * n_frames)

        crafter = SlidingWindowAudioSlicer(frame_length, frame_length // 2)
        crafted_chunks = crafter.craft(signal_orig, 0)

        assert len(crafted_chunks) == n_frames * 2 - 1

    def test_sliding_window_stereo(self):
        n_frames = 100
        frame_length = 2048
        signal_orig = np.random.randn(2, frame_length * n_frames)

        crafter = SlidingWindowAudioSlicer(frame_length, frame_length // 2)
        crafted_chunks = crafter.craft(signal_orig, 0)

        assert len(crafted_chunks) == (n_frames * 2 - 1) * 2



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders import BaseAudioEncoder
from jina.executors.encoders.frameworks import BaseTorchEncoder
from jina.helper import cached_property


class Wav2VecSpeechEncoder(BaseTorchEncoder, BaseAudioEncoder):
    """
    :class:`Wav2VecSpeechEncoder` is a speech encoder based on `wav2vec` - unsupervised pre-training for speech
    recognition presented and implemented by Facebook: https://github.com/pytorch/fairseq/tree/master/examples/wav2vec
    :class:`Wav2VecSpeechEncoder` uses a pre-trained model to encode an audio signal from a `Batch x Signal Length`
    ndarray into a `Batch x Concatenated Features` ndarray.
    """

    def __init__(self, model_path: str, input_sample_rate: int = 22050, *args, **kwargs):
        """
        Wav2vec model produces a representation for each time step at a rate of 100 Hz.

        :param model_path: the path of the pre-trained model. The pre-trained model can be downloaded at
            https://github.com/pytorch/fairseq/blob/master/examples/Wav2Vec/README.md#pre-trained-models
        :param input_sample_rate: input sampling rate in Hz (22050 by default)
        """
        super().__init__(*args, **kwargs)
        self.model_path = model_path
        self.input_sample_rate = input_sample_rate

    def post_init(self):
        super().post_init()
        import torch
        from fairseq.models.wav2vec import Wav2VecModel
        cp = torch.load(self.model_path, map_location=torch.device('cpu'))
        self.model = Wav2VecModel.build_model(cp['args'], task=None)
        self.model.load_state_dict(cp['model'])
        self.model.eval()
        self.to_device(self.model)
        self._tensor_func = torch.tensor

    @batching
    @as_ndarray
    def encode(self, data: np.ndarray, *args, **kwargs) -> np.ndarray:
        """
        Resamples the input audio signal to 16kHz, segments the resampled signal of each Chunk into wav2vec frames,
        encodes the frames and concatenates Chunk frame embeddings into a single Chunk embedding.

        :param data: a `Batch x Signal Length` ndarray, where `Signal Length` is a number of samples
        :return: a `Batch x Concatenated Features` ndarray, where `Concatinated Features` is a 512-dimensional feature
        vector times the number of the wav2vec frames
        """
        assert data.shape[1] >= 465, 'the signal must have at least 465 samples'
        from librosa import resample
        embeds = []
        with self.session():
            for chunk_data in data:
                resampled_signal = resample(chunk_data, self.input_sample_rate, 16000)
                signal_tensor = self.array2tensor(resampled_signal.reshape(1, -1))
                features = self.model.feature_extractor(signal_tensor)
                embed_tensor = self.model.feature_aggregator(features)[0]
                chunk_embed = self.tensor2array(embed_tensor).T.flatten()
                embeds.append(chunk_embed)
        return embeds

    def array2tensor(self, array):
        tensor = self._tensor_func(array)
        return tensor.cuda() if self.on_gpu else tensor

    def tensor2array(self, tensor):
        return tensor.cuda().numpy() if self.on_gpu else tensor.numpy()

    @cached_property
    def session(self):
        return self.get_session()

    def get_session(self):
        from torch import no_grad
        return no_grad



from typing import Dict, List

from jina.executors.crafters import BaseSegmenter


class JiebaSegmenter(BaseSegmenter):
    """
    :class:`JiebaSegmenter` split the chinese text on the doc-level into words on the chunk-level with `jieba`.
    """

    def __init__(self, mode: str = 'accurate', *args, **kwargs):
        """

        :param mode: the jieba cut mode, accurate, all, search. default accurate
        """
        super().__init__(*args, **kwargs)
        if mode not in ('accurate', 'all', 'search'):
            raise ValueError('you must choose one of modes to cut the text: accurate, all, search.')
        self.mode = mode

    def craft(self, text: str, *args, **kwargs) -> List[Dict]:
        """
        Split the chinese text into words
        :param text: the raw text
        :return: a list of chunk dicts
        """
        import jieba
        if self.mode == 'search':
            words = jieba.cut_for_search(text)
        elif self.mode == 'all':
            words = jieba.cut(text, cut_all=True)
        else:
            words = jieba.cut(text)

        chunks = []
        for idx, word in enumerate(words):
            chunks.append(
                dict(text=word, offset=idx, weight=1.0))

        return chunks



import os
import numpy as np
from PIL import Image

from unittest.mock import patch
from .. import TorchObjectDetectionSegmenter

cur_dir = os.path.dirname(os.path.abspath(__file__))

def create_test_image(output_fn, size_width=50, size_height=50):
    image = Image.new('RGB', size=(size_width, size_height), color=(155, 0, 0))
    with open(output_fn, "wb") as f:
        image.save(f, 'jpeg')

def create_random_img_array(img_height, img_width):
    import numpy as np
    return np.random.randint(0, 256, (img_height, img_width, 3))


class MockModel:
    def __init__(self):
        pass

    def __call__(self, input_ids, *args, **kwargs):
        import torch
        bbox_1 = [10, 15, 30, 40]
        bbox_2 = [-1, -1, -1, -1]
        bbox_3 = [20, 10, 30, 40]
        score_1 = 0.91
        score_2 = 0.87
        score_3 = 0.909
        label_1 = 1
        label_2 = 2
        label_3 = 3
        predictions = [{
            'boxes': torch.Tensor([bbox_1, bbox_2, bbox_3]),
            'scores': torch.Tensor([score_1, score_2, score_3]),
            'labels': torch.Tensor([label_1, label_2, label_3])
        }]
        return predictions

    def eval(self):
        return self

    def to(self, device):
        return self


def test_encoding_mock_model_results():
    import torchvision.models.detection as detection_models
    img_array = create_random_img_array(128, 64)
    img_array = img_array / 255
    with patch.object(detection_models, 'fasterrcnn_resnet50_fpn', return_value=MockModel()):
        crafter = TorchObjectDetectionSegmenter(channel_axis=-1, confidence_threshold=0.9,
                                                label_name_map={0: 'zero',
                                                                1: 'one',
                                                                2: 'two',
                                                                3: 'three'})
        chunks = crafter.craft(img_array)
        assert len(chunks) == 2
        assert chunks[0]['blob'].shape == (25, 20, 3)
        assert chunks[0]['location'] == (15, 10)
        assert chunks[0]['meta_info'].decode() == 'one'

        assert chunks[1]['blob'].shape == (30, 10, 3)
        assert chunks[1]['location'] == (10, 20)
        assert chunks[1]['meta_info'].decode() == 'three'


def test_encoding_fasterrcnn_results():
    img_array = create_random_img_array(128, 64)
    img_array = img_array / 255
    crafter = TorchObjectDetectionSegmenter(channel_axis=-1, confidence_threshold=0.98)
    chunks = crafter.craft(img_array)
    assert len(chunks) == 0


def test_encoding_fasterrcnn_results_real_image():
    """
    Credit for the image used in this test: 
    Photo by <a href="/photographer/createsima-47728">createsima</a> from <a href="https://freeimages.com/">FreeImages</a>
    https://www.freeimages.com/photo/cars-1407390
    TorchObjectDete@29513[I]:detected person with confidence 0.9911105632781982 at position (541, 992) and size (67, 24)
    TorchObjectDete@29513[I]:detected car with confidence 0.9843265414237976 at position (496, 201) and size (104, 161)
    TorchObjectDete@29513[I]:detected car with confidence 0.9835659861564636 at position (524, 574) and size (77, 131)
    TorchObjectDete@29513[I]:detected person with confidence 0.9795390367507935 at position (539, 969) and size (66, 27)
    TorchObjectDete@29513[I]:detected person with confidence 0.9787288904190063 at position (530, 934) and size (74, 18)
    TorchObjectDete@29513[I]:detected car with confidence 0.9717466831207275 at position (517, 377) and size (82, 154)
    TorchObjectDete@29513[I]:detected person with confidence 0.9682216048240662 at position (532, 919) and size (70, 19)
    TorchObjectDete@29513[I]:detected truck with confidence 0.964297354221344 at position (498, 702) and size (106, 169)
    TorchObjectDete@29513[I]:detected car with confidence 0.9574888944625854 at position (522, 1046) and size (88, 164)
    TorchObjectDete@29513[I]:detected person with confidence 0.9304793477058411 at position (536, 962) and size (70, 17)
    """
    path = os.path.join(cur_dir, 'imgs/cars.jpg')
    img = Image.open(path)
    img = img.convert('RGB')
    img_array = np.array(img).astype('float32') / 255
    crafter = TorchObjectDetectionSegmenter(channel_axis=-1, confidence_threshold=0.9)
    chunks = crafter.craft(img_array)
    assert len(chunks) == 10
    assert chunks[0]['meta_info'].decode() == 'person'
    img = Image.open(os.path.join(cur_dir, 'imgs/faster_rcnn/person-0.png'))
    assert chunks[0]['location'] == (541, 992)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[0]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (67, 24, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[1]['meta_info'].decode() == 'car'
    img = Image.open(os.path.join(cur_dir, 'imgs/faster_rcnn/car-1.png'))
    assert chunks[1]['location'] == (496, 201)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[1]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (104, 161, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[2]['meta_info'].decode() == 'car'
    img = Image.open(os.path.join(cur_dir, 'imgs/faster_rcnn/car-2.png'))
    assert chunks[2]['location'] == (524, 574)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[2]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (77, 131, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[3]['meta_info'].decode() == 'person'
    img = Image.open(os.path.join(cur_dir, 'imgs/faster_rcnn/person-3.png'))
    assert chunks[3]['location'] == (539, 969)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[3]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (66, 27, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[4]['meta_info'].decode() == 'person'
    img = Image.open(os.path.join(cur_dir, 'imgs/faster_rcnn/person-4.png'))
    assert chunks[4]['location'] == (530, 934)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[4]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (74, 18, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[5]['meta_info'].decode() == 'car'
    img = Image.open(os.path.join(cur_dir, 'imgs/faster_rcnn/car-5.png'))
    assert chunks[5]['location'] == (517, 377)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[5]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (82, 154, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[6]['meta_info'].decode() == 'person'
    img = Image.open(os.path.join(cur_dir, 'imgs/faster_rcnn/person-6.png'))
    assert chunks[6]['location'] == (532, 919)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[6]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (70, 19, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    # it missclassifies as truck (but is a fairly big car)
    assert chunks[7]['meta_info'].decode() == 'truck'
    img = Image.open(os.path.join(cur_dir, 'imgs/faster_rcnn/car-7.png'))
    assert chunks[7]['location'] == (498, 702)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[7]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (106, 169, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[8]['meta_info'].decode() == 'car'
    img = Image.open(os.path.join(cur_dir, 'imgs/faster_rcnn/car-8.png'))
    assert chunks[8]['location'] == (522, 1046)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[8]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (88, 164, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[9]['meta_info'].decode() == 'person'
    img = Image.open(os.path.join(cur_dir, 'imgs/faster_rcnn/person-9.png'))
    assert chunks[9]['location'] == (536, 962)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[9]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (70, 17, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)


def test_encoding_maskrcnn_results():
    img_array = create_random_img_array(128, 64)
    img_array = img_array / 255
    crafter = TorchObjectDetectionSegmenter(model_name='maskrcnn_resnet50_fpn',
                                            channel_axis=-1, confidence_threshold=0.98)
    chunks = crafter.craft(img_array)
    assert len(chunks) == 0


def test_encoding_maskrcnn_results_real_image():
    """
    Credit for the image used in this test:
    Photo by <a href="/photographer/createsima-47728">createsima</a> from <a href="https://freeimages.com/">FreeImages</a>
    https://www.freeimages.com/photo/cars-1407390
    TorchObjectDete@31595[I]:detected car with confidence 0.996136486530304 at position (518, 379) and size (85, 152)
    TorchObjectDete@31595[I]:detected car with confidence 0.9923352599143982 at position (527, 572) and size (74, 134)
    TorchObjectDete@31595[I]:detected person with confidence 0.9859431982040405 at position (541, 993) and size (66, 23)
    TorchObjectDete@31595[I]:detected person with confidence 0.9840929508209229 at position (531, 934) and size (74, 19)
    TorchObjectDete@31595[I]:detected car with confidence 0.9836736917495728 at position (499, 196) and size (103, 165)
    TorchObjectDete@31595[I]:detected person with confidence 0.9716113805770874 at position (532, 917) and size (71, 20)
    TorchObjectDete@31595[I]:detected person with confidence 0.9690250754356384 at position (539, 968) and size (68, 27)
    TorchObjectDete@31595[I]:detected truck with confidence 0.9677107334136963 at position (499, 700) and size (107, 167)
    TorchObjectDete@31595[I]:detected car with confidence 0.9577637314796448 at position (534, 142) and size (64, 65)
    TorchObjectDete@31595[I]:detected car with confidence 0.9379956126213074 at position (521, 1037) and size (90, 175)

    """
    path = os.path.join(cur_dir, 'imgs/cars.jpg')
    img = Image.open(path)
    img = img.convert('RGB')
    img_array = np.array(img).astype('float32') / 255
    crafter = TorchObjectDetectionSegmenter(model_name='maskrcnn_resnet50_fpn',
                                            channel_axis=-1, confidence_threshold=0.9)
    chunks = crafter.craft(img_array)

    assert len(chunks) == 10
    assert chunks[0]['meta_info'].decode() == 'car'
    img = Image.open(os.path.join(cur_dir, 'imgs/mask_rcnn/car-0.png'))
    assert chunks[0]['location'] == (518, 379)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[0]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (85, 152, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[1]['meta_info'].decode() == 'car'
    img = Image.open(os.path.join(cur_dir, 'imgs/mask_rcnn/car-1.png'))
    assert chunks[1]['location'] == (527, 572)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[1]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (74, 134, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[2]['meta_info'].decode() == 'person'
    img = Image.open(os.path.join(cur_dir, 'imgs/mask_rcnn/person-2.png'))
    assert chunks[2]['location'] == (541, 993)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[2]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (66, 23, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[3]['meta_info'].decode() == 'person'
    img = Image.open(os.path.join(cur_dir, 'imgs/mask_rcnn/person-3.png'))
    assert chunks[3]['location'] == (531, 934)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[3]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (74, 19, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[4]['meta_info'].decode() == 'car'
    img = Image.open(os.path.join(cur_dir, 'imgs/mask_rcnn/car-4.png'))
    assert chunks[4]['location'] == (499, 196)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[4]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (103, 165, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[5]['meta_info'].decode() == 'person'
    img = Image.open(os.path.join(cur_dir, 'imgs/mask_rcnn/person-5.png'))
    assert chunks[5]['location'] == (532, 917)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[5]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (71, 20, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[6]['meta_info'].decode() == 'person'
    img = Image.open(os.path.join(cur_dir, 'imgs/mask_rcnn/person-6.png'))
    assert chunks[6]['location'] == (539, 968)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[6]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (68, 27, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    # it missclassifies as truck (but is a fairly big car)
    assert chunks[7]['meta_info'].decode() == 'truck'
    img = Image.open(os.path.join(cur_dir, 'imgs/mask_rcnn/car-7.png'))
    assert chunks[7]['location'] == (499, 700)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[7]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (107, 167, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[8]['meta_info'].decode() == 'car'
    img = Image.open(os.path.join(cur_dir, 'imgs/mask_rcnn/car-8.png'))
    assert chunks[8]['location'] == (534, 142)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[8]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (64, 65, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)

    assert chunks[9]['meta_info'].decode() == 'car'
    img = Image.open(os.path.join(cur_dir, 'imgs/mask_rcnn/car-9.png'))
    assert chunks[9]['location'] == (521, 1037)
    # check that the shape of retrieved is the same as the expected image (was computed and stored once)
    blob = chunks[9]['blob']
    assert (blob.shape[1], blob.shape[0]) == img.size
    assert blob.shape == (90, 175, 3)
    array = np.array(img)
    np.testing.assert_array_almost_equal(blob, array)



import numpy as np
from PIL import Image

from .. import CenterImageCropper

def create_random_img_array(img_height, img_width):
    import numpy as np
    return np.random.randint(0, 256, (img_height, img_width, 3))


def test_center_crop():
    img_size = 217
    img_array = create_random_img_array(img_size, img_size)
    width = 30
    height = 20
    output_dim = (height, width)
    crafter = CenterImageCropper(output_dim)
    crafted_doc = crafter.craft(img_array)
    assert crafted_doc['blob'].shape == (height, width, 3)
    # int((img_size - output_dim) / 2)
    crop = Image.fromarray(np.uint8(crafted_doc['blob']))
    crop_width, crop_height = crop.size
    assert crop_width == width
    assert crop_height == height
    (top, left) = (98, 93)
    assert crafted_doc['location'] == (top, left)



from jina.flow import Flow

from .. import Sentencizer


class MyTestCase:
    def test_sentencier_en(self):
        sentencizer = Sentencizer()
        text = 'It is a sunny day!!!! When Andy comes back, we are going to the zoo.'
        crafted_chunk_list = sentencizer.craft(text, 0)
        assert len(crafted_chunk_list) == 2

    def test_sentencier_en_new_lines(self):
        """
        New lines are also considered as a separator.
        """
        sentencizer = Sentencizer()
        text = 'It is a sunny day!!!! When Andy comes back,\n' \
               'we are going to the zoo.'
        crafted_chunk_list = sentencizer.craft(text, 0)
        assert len(crafted_chunk_list) == 3

    def test_sentencier_en_float_numbers(self):
        """
        Separators in float numbers, URLs, emails, abbreviations (like 'Mr.')
        are not taking into account.
        """
        sentencizer = Sentencizer()
        text = 'With a 0.99 probability this sentence will be ' \
               'tokenized in 2 sentences.'
        crafted_chunk_list = sentencizer.craft(text, 0)
        assert len(crafted_chunk_list) == 2

    def test_sentencier_en_trim_spaces(self):
        """
        Trimming all spaces at the beginning an end of the chunks.
        Keeping extra spaces inside chunks.
        Ignoring chunks with only spaces.
        """
        sentencizer = Sentencizer()
        text = '  This ,  text is...  . Amazing !!'
        chunks = [i['text'] for i in sentencizer.craft(text, 0)]
        locs = [i['location'] for i in sentencizer.craft(text, 0)]
        assert chunks, ["This ,  text is..." == "Amazing"]
        assert text[locs[0][0]:locs[0][1]], '  This  ==   text is...'
        assert text[locs[1][0]:locs[1][1]] == ' Amazing'

        def validate(req):
            assert req.docs[0].chunks[0].text, 'This  ==   text is...'
            assert req.docs[0].chunks[1].text == 'Amazing'

        f = Flow().add(uses='!Sentencizer')
        with f:
            f.index_lines(['  This ,  text is...  . Amazing !!'], output_fn=validate, callback_on_body=True)

    def test_sentencier_cn(self):
        sentencizer = Sentencizer()
        text = '今天是个大晴天！安迪回来以后，我们准备去动物园。'
        crafted_chunk_list = sentencizer.craft(text, 0)
        # Sentencizer does not work for chinese because string.printable does not contain Chinese characters
        assert len(crafted_chunk_list) == 0



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import atexit
import multiprocessing

__sse_queue__ = multiprocessing.Queue()  #: the global sse log queue
__profile_queue__ = multiprocessing.Queue()  #: the global profile log queue
__log_queue__ = multiprocessing.Queue()  #: the global log queue


def clear_queue():
    """Clear the log queue and profile queue when the program exit

    This is only used when server-side event (SSE) logging is turned on.
    """
    try:
        while not __sse_queue__.empty():
            __sse_queue__.get_nowait()

        while not __profile_queue__.empty():
            __profile_queue__.get_nowait()

        while not __log_queue__.empty():
            __log_queue__.get_nowait()
    except:
        # let's ignore this for a peaceful ending
        pass


atexit.register(clear_queue)  #: clear the log queue when exit



from .. import JiebaSegmenter


class TestClass:
    def test_jieba_crafter(self):
        jieba_crafter = JiebaSegmenter(mode='accurate')
        text = '今天是个大晴天！安迪回来以后，我们准备去动物园。'
        crafted_chunk_list = jieba_crafter.craft(text, 0)
        assert len(crafted_chunk_list) == 14



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from jina.executors.rankers import Chunk2DocRanker


class BiMatchRanker(Chunk2DocRanker):
    """The :class:`BiMatchRanker` counts the best chunk-hit from both query and doc perspective.

    .. warning:: Here we suppose that the smaller chunk score means the more similar.
    """
    required_keys = {'length'}
    D_MISS = 2000  # cost of a non-match chunk, used for normalization

    def _get_score(self, match_idx, query_chunk_meta, match_chunk_meta, *args, **kwargs):
        s1 = self._directional_score(match_idx, match_chunk_meta, col=self.col_chunk_id)
        s2 = self._directional_score(match_idx, query_chunk_meta, col=self.col_query_chunk_id)
        return self.get_doc_id(match_idx), (s1 + s2) / 2.

    def _directional_score(self, g, chunk_meta, col):
        # col = self.col_chunk_id, from matched_chunk aspect
        # col = self.col_query_chunk_id, from query chunk aspect
        # group by col
        _groups = self._group_by(g, col)
        # take the best match from each group
        _groups_best = np.stack([gg[gg[:, -1].argsort()][0] for gg in _groups])
        # doc total length
        _c = chunk_meta[_groups_best[0, col]]['length']
        # hit chunks
        _h = _groups_best.shape[0]
        # hit distance
        sum_d_hit = np.sum(_groups_best[:, -1])
        # all hit => 0, all_miss => 1
        return 1 - (sum_d_hit + self.D_MISS * (_c - _h)) / (self.D_MISS * _c)



import numpy as np

from .. import AudioNormalizer


class TestClass:
    def test_audionormalizer(self):
        """here is my test code

        https://docs.pytest.org/en/stable/getting-started.html#create-your-first-test
        """
        signal_orig = np.random.randn(2, 31337)

        crafter = AudioNormalizer()
        crafted_doc = crafter.craft(signal_orig, 0)

        signal_norm = crafted_doc["blob"]
        assert signal_norm.shape == signal_orig.shape
        assert np.min(signal_norm) == -1
        assert np.max(signal_norm) == 1



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import jina_pb2 as jina__pb2


class JinaRPCStub(object):
  """*
  jina gRPC service.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.Call = channel.stream_stream(
        '/jina.JinaRPC/Call',
        request_serializer=jina__pb2.Request.SerializeToString,
        response_deserializer=jina__pb2.Request.FromString,
        )
    self.CallUnary = channel.unary_unary(
        '/jina.JinaRPC/CallUnary',
        request_serializer=jina__pb2.Request.SerializeToString,
        response_deserializer=jina__pb2.Request.FromString,
        )
    self.Spawn = channel.unary_stream(
        '/jina.JinaRPC/Spawn',
        request_serializer=jina__pb2.SpawnRequest.SerializeToString,
        response_deserializer=jina__pb2.SpawnRequest.FromString,
        )


class JinaRPCServicer(object):
  """*
  jina gRPC service.
  """

  def Call(self, request_iterator, context):
    """Pass in a Request and a filled Request with matches will be returned.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CallUnary(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Spawn(self, request, context):
    """Pass in a Request and a filled Request will be returned.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_JinaRPCServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'Call': grpc.stream_stream_rpc_method_handler(
          servicer.Call,
          request_deserializer=jina__pb2.Request.FromString,
          response_serializer=jina__pb2.Request.SerializeToString,
      ),
      'CallUnary': grpc.unary_unary_rpc_method_handler(
          servicer.CallUnary,
          request_deserializer=jina__pb2.Request.FromString,
          response_serializer=jina__pb2.Request.SerializeToString,
      ),
      'Spawn': grpc.unary_stream_rpc_method_handler(
          servicer.Spawn,
          request_deserializer=jina__pb2.SpawnRequest.FromString,
          response_serializer=jina__pb2.SpawnRequest.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'jina.JinaRPC', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os

from . import BaseEncoder
from ..devices import OnnxDevice, PaddleDevice, TorchDevice, TFDevice, MindsporeDevice
from ...helper import is_url
from ...helper import cached_property


# mixin classes go first, base classes are read from right to left.
class BaseOnnxEncoder(OnnxDevice, BaseEncoder):
    def __init__(self, output_feature: str, model_path: str = None, *args, **kwargs):
        """

        :param output_feature: the name of the layer for feature extraction.
        :param model_path: the path of the model in the format of `.onnx`. Check a list of available pretrained
            models at https://github.com/onnx/models#image_classification and download the git LFS to your local path.
            The ``model_path`` is the local path of the ``.onnx`` file, e.g. ``/tmp/onnx/mobilenetv2-1.0.onnx``.
        """
        super().__init__(*args, **kwargs)
        self.outputs_name = output_feature
        self.raw_model_path = model_path

    def post_init(self):
        """
        Load the model from the `.onnx` file and add outputs for the selected layer, i.e. ``outputs_name``. The modified
             models is saved at `tmp_model_path`.
        """
        import onnxruntime
        super().post_init()
        self.model_name = self.raw_model_path.split('/')[-1]
        self.tmp_model_path = self.get_file_from_workspace(f'{self.model_name}.tmp')
        if is_url(self.raw_model_path):
            import urllib.request
            download_path, *_ = urllib.request.urlretrieve(self.raw_model_path)
            self.raw_model_path = download_path
            self.logger.info(f'download the model at {self.raw_model_path}')
        if not os.path.exists(self.tmp_model_path):
            self._append_outputs(self.raw_model_path, self.outputs_name, self.tmp_model_path)
            self.logger.info(f'save the model with outputs [{self.outputs_name}] at {self.tmp_model_path}')
        self.model = onnxruntime.InferenceSession(self.tmp_model_path, None)
        self.inputs_name = self.model.get_inputs()[0].name
        self._device = None
        self.to_device(self.model)

    @staticmethod
    def _append_outputs(input_fn, outputs_name_to_append, output_fn):
        import onnx
        model = onnx.load(input_fn)
        feature_map = onnx.helper.ValueInfoProto()
        feature_map.name = outputs_name_to_append
        model.graph.output.append(feature_map)
        onnx.save(model, output_fn)


class BaseTFEncoder(TFDevice, BaseEncoder):
    pass


class BaseTorchEncoder(TorchDevice, BaseEncoder):
    pass


class BasePaddleEncoder(PaddleDevice, BaseEncoder):
    pass


class BaseMindsporeEncoder(MindsporeDevice, BaseEncoder):
    """
    :class:`BaseMindsporeEncoder` is the base class for implementing Encoders with models from `mindspore`.

    To implement your own executor with the :mod:`mindspore` lilbrary,

    .. highlight:: python
    .. code-block:: python
        import mindspore.nn as nn

        class YourAwesomeModel(nn.Cell):
            def __init__(self):
                ...

            def construct(self, x):
                ...

        class YourAwesomeEncoder(BaseMindsporeEncoder):
            def encode(self, data, *args, **kwargs):
                from mindspore import Tensor
                return self.model(Tensor(data)).asnumpy()

            def get_model(self):
                return YourAwesomeModel()

    """
    def __init__(self, model_path: str, *args, **kwargs):
        """

        :param model_path: the path of the model's checkpoint.
        """
        super().__init__(*args, **kwargs)
        self.model_path = model_path

    def post_init(self):
        """
        Load the model from the `.ckpt` checkpoint.
        """
        from mindspore.train.serialization import load_checkpoint, load_param_into_net
        super().post_init()
        self.to_device()
        _param_dict = load_checkpoint(ckpt_file_name=self.model_path)
        load_param_into_net(self.model, _param_dict)

    @cached_property
    def model(self):
        return self.get_model()

    def get_model(self):
        raise NotImplemented('the model is not implemented')



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple

import numpy as np
from jina.executors.devices import FaissDevice
from jina.executors.indexers.vector import BaseNumpyIndexer


class FaissIndexer(FaissDevice, BaseNumpyIndexer):
    """Faiss powered vector indexer

    For more information about the Faiss supported parameters and installation problems, please consult:
        - https://github.com/facebookresearch/faiss

    .. note::
        Faiss package dependency is only required at the query time.
    """

    def __init__(self, index_key: str, train_filepath: str = None,
                 distance: str = 'l2', nprobe: int = 1, *args, **kwargs):
        """
        Initialize an Faiss Indexer

        :param index_key: index type supported by ``faiss.index_factory``
        :param train_filepath: the training data file path, e.g ``faiss.tgz`` or `faiss.npy`. The data file is expected
            to be either `.npy` file from `numpy.save()` or a `.tgz` file from `NumpyIndexer`.
        :param distance: 'l2' or 'inner_product' accepted. Determines which distances to optimize by FAISS
        :param nprobe: Number of clusters to consider at search time.

        .. highlight:: python
        .. code-block:: python
            # generate a training file in `.tgz`
            import gzip
            import numpy as np
            from jina.executors.indexers.vector.faiss import FaissIndexer

            train_filepath = 'faiss_train.tgz'
            train_data = np.random.rand(10000, 128)
            with gzip.open(train_filepath, 'wb', compresslevel=1) as f:
                f.write(train_data.astype('float32'))
            indexer = FaissIndexer('PCA64,FLAT', train_filepath)

            # generate a training file in `.npy`
            train_filepath = 'faiss_train'
            np.save(train_filepath, train_data)
            indexer = FaissIndexer('PCA64,FLAT', train_filepath)
        """
        super().__init__(*args, **kwargs)
        self.index_key = index_key
        self.train_filepath = train_filepath
        self.distance = distance
        self.nprobe = nprobe

    def build_advanced_index(self, vecs: 'np.ndarray'):
        """Load all vectors (in numpy ndarray) into Faiss indexers """
        import faiss
        metric = faiss.METRIC_L2
        if self.distance == 'inner_product':
            metric = faiss.METRIC_INNER_PRODUCT
        if self.distance not in {'inner_product', 'l2'}:
            self.logger.warning('Invalid distance metric for Faiss index construction. Defaulting to l2 distance')

        index = self.to_device(index=faiss.index_factory(self.num_dim, self.index_key, metric))

        if not self.is_trained:
            train_data = self._load_training_data(self.train_filepath)
            if train_data is None:
                self.logger.warning('loading training data failed.')
                return None
            self.train(index, train_data)
        index.add(vecs.astype('float32'))
        index.nprobe = self.nprobe
        return index

    def query(self, keys: 'np.ndarray', top_k: int, *args, **kwargs) -> Tuple['np.ndarray', 'np.ndarray']:
        dist, ids = self.query_handler.search(keys, top_k)
        return self.int2ext_key[ids], dist

    def train(self, index, data: 'np.ndarray', *args, **kwargs) -> None:
        _num_samples, _num_dim = data.shape
        if not self.num_dim:
            self.num_dim = _num_dim
        if self.num_dim != _num_dim:
            raise ValueError('training data should have the same number of features as the index, {} != {}'.format(
                self.num_dim, _num_dim))
        index.train(data)

    def _load_training_data(self, train_filepath: str) -> 'np.ndarray':
        result = None
        try:
            result = self._load_gzip(train_filepath)
            if result is not None:
                return result
        except OSError as e:
            self.logger.info('not a gzippped file, {}'.format(e))

        try:
            result = np.load(train_filepath)
            if isinstance(result, np.lib.npyio.NpzFile):
                self.logger.warning('.npz format is not supported. Please save the array in .npy format.')
                result = None
        except Exception as e:
            self.logger.error('loading training data failed, filepath={}, {}'.format(train_filepath, e))
        return result



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from collections import defaultdict
from typing import Dict, List, Iterable

import numpy as np

from . import BaseRecursiveDriver
from .helper import pb2array, array2pb
from ..excepts import NoExplicitMessage
from ..proto import jina_pb2


class ReduceDriver(BaseRecursiveDriver):
    """:class:`ReduceDriver` merges envelope of all requests """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._prev_requests = None
        self._prev_messages = None
        self._pending_msgs = defaultdict(list)  # type: Dict[str, List]

    @property
    def prev_reqs(self) -> List['jina_pb2.Request']:
        """Get all previous requests that has the same ``request_id``, shortcut to ``self.pea.prev_requests``

        This returns ``None`` when ``num_part=1``.
        """
        return self._prev_requests

    @property
    def prev_reqs_exclude_last(self) -> List['jina_pb2.Request']:
        """Get all previous requests but excluding the current request (last received request)
        """
        return self._prev_requests[:-1]

    @property
    def prev_msgs(self) -> List['jina_pb2.Message']:
        """Get all previous messages that has the same ``request_id``, shortcut to ``self.pea.prev_messages``

        This returns ``None`` when ``num_part=1``.
        """
        return self._prev_messages

    def __call__(self, *args, **kwargs):
        if self.envelope.num_part[-1] > 1:
            req_id = self.envelope.request_id
            self._pending_msgs[req_id].append(self.msg)
            num_req = len(self._pending_msgs[req_id])

            self.logger.info(f'collected {num_req}/{self.envelope.num_part[-1]} parts of {type(self.req).__name__}')

            if num_req == self.envelope.num_part[-1]:
                self._prev_messages = self._pending_msgs.pop(req_id)
                self._prev_requests = [getattr(v.request, v.request.WhichOneof('body')) for v in
                                       self._prev_messages]
            else:
                raise NoExplicitMessage

            if self.req.docs:
                self.reduce(*args, **kwargs)

            # merge envelope
            routes = {(r.pod + r.pod_id): r for m in self.prev_msgs for r in m.envelope.routes}
            self.msg.envelope.ClearField('routes')
            self.msg.envelope.routes.extend(
                sorted(routes.values(), key=lambda x: (x.start_time.seconds, x.start_time.nanos)))
            self.envelope.num_part.pop(-1)

    def reduce(self, *args, **kwargs):
        """ Reduce the message from all requests by merging their envelopes
        """
        # take unique routes by service identity


class ReduceAllDriver(ReduceDriver):
    """:class:`ReduceAllDriver` merges chunks/matches from all requests, recursively.

    .. note::

        It uses the last request as a reference.
    """

    def _apply(self, doc: 'jina_pb2.Document', context_doc: 'jina_pb2.Document', field: str, *args, **kwargs):
        if doc.id not in self.doc_pointers:
            self.doc_pointers[doc.id] = doc

    def reduce(self, *args, **kwargs):
        self._is_apply, self._is_apply_all = True, False
        # use docs in the last request to set the pointers
        self.doc_pointers = {}
        self._traverse_apply(self.req.docs, *args, **kwargs)

        self._is_apply, self._is_apply_all = False, True

        # traverse apply on ALL previous requests collected
        for r in self.prev_reqs_exclude_last:
            self._traverse_apply(r.docs, *args, **kwargs)

    def _apply_all(self, docs: Iterable['jina_pb2.Document'], context_doc: 'jina_pb2.Document', field: str, *args,
                   **kwargs):
        if context_doc:
            getattr(self.doc_pointers[context_doc.id], field).extend(docs)


class ConcatEmbedDriver(ReduceDriver):
    def _apply(self, doc: 'jina_pb2.Document', context_doc: 'jina_pb2.Document', field: str, *args, **kwargs):
        if doc.id not in self.doc_pointers:
            self.doc_pointers[doc.id] = [pb2array(doc.embedding)]
        else:
            self.doc_pointers[doc.id].append(pb2array(doc.embedding))

    def _apply_post(self, doc: 'jina_pb2.Document', *args, **kwargs):
        doc.embedding.CopyFrom(array2pb(np.concatenate(self.doc_pointers[doc.id], axis=0)))

    def reduce(self, *args, **kwargs):
        # use docs in the last request to set the pointers
        self.doc_pointers = {}
        self._traverse_apply(self.req.docs, *args, **kwargs)

        # traverse apply on ALL previous requests collected
        for r in self.prev_reqs_exclude_last:
            self._traverse_apply(r.docs, *args, **kwargs)

        # update embedding
        self._apply = self._apply_post
        self._traverse_apply(self.req.docs, *args, **kwargs)



import os
import shutil
import time
import pytest

import numpy as np
from milvus import Milvus

from .. import MilvusIndexer
from ..MilvusDBHandler import MilvusDBHandler

cur_dir = os.path.dirname(os.path.abspath(__file__))

port = 19530
host = '127.0.0.1'
img_name = 'milvusdb/milvus:0.10.0-cpu-d061620-5f3c00'
host_milvus_tmp = os.path.join(cur_dir, 'milvus_tmp')


def rm_files(file_paths):
    for file_path in file_paths:
        if os.path.exists(file_path):
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path, ignore_errors=False, onerror=None)


def create_collection(collection_name):
    client = Milvus(host, str(port))

    status, ok = client.has_collection(collection_name)
    if not ok:
        param = {
            'collection_name': collection_name,
            'dimension': 3,
        }
        client.create_collection(param)
    client.close()


def docker_run():
    import docker
    conf_tmp = os.path.join(host_milvus_tmp, 'conf')
    os.makedirs(host_milvus_tmp, exist_ok=True)
    os.makedirs(conf_tmp, exist_ok=True)
    shutil.copy(os.path.join(cur_dir, 'server_config.yaml'), conf_tmp)
    client = docker.from_env()
    client.images.pull(img_name)

    bind_volumes = {
        os.path.join(host_milvus_tmp, 'db'): {'bind': '/var/lib/milvusdb/db', 'mode': 'rw'},
        os.path.join(host_milvus_tmp, 'conf'): {'bind': '/var/lib/milvusdb/conf', 'mode': 'rw'},
        os.path.join(host_milvus_tmp, 'logs'): {'bind': '/var/lib/milvusdb/logs', 'mode': 'rw'},
        os.path.join(host_milvus_tmp, 'wal'): {'bind': '/var/lib/milvusdb/wal', 'mode': 'rw'}
    }

    container = client.containers.run(img_name, name='milvus_test_image',
                                      volumes=bind_volumes, detach=True, auto_remove=True,
                                      ports={f'{port}/tcp': f'{port}', '19121/tcp': '19121'},
                                      network_mode='host')
    client.close()
    return container


def docker_clean(container):
    container.stop()


def setup_tear_down_decorator(func):
    def wrapper():
        container = docker_run()
        func()
        docker_clean(container)
        time.sleep(2)
        rm_files([host_milvus_tmp])

    return wrapper


@pytest.mark.skip
@setup_tear_down_decorator
def test_milvusdbhandler_simple():
    collection_name = 'simple_milvus'
    create_collection(collection_name)

    vectors = np.array([[1, 1, 1],
                        [10, 10, 10],
                        [100, 100, 100],
                        [1000, 1000, 1000]])
    keys = np.array([0, 1, 2, 3]).reshape(-1, 1)
    with MilvusDBHandler(host, port, collection_name) as db:
        db.insert(keys, vectors)
        dist, idx = db.search(vectors, 2)
        dist = np.array(dist)
        idx = np.array(idx)
        assert idx.shape == dist.shape
        assert idx.shape == (4, 2)
        np.testing.assert_equal(idx, np.array([[0, 1], [1, 0], [2, 1], [3, 2]]))


@pytest.mark.skip
@setup_tear_down_decorator
def test_milvusdbhandler_build():
    collection_name = 'build_milvus'
    create_collection(collection_name)

    vectors = np.array([[1, 1, 1],
                        [10, 10, 10],
                        [100, 100, 100],
                        [1000, 1000, 1000]])
    keys = np.array([0, 1, 2, 3]).reshape(-1, 1)
    with MilvusDBHandler(host, port, collection_name) as db:
        db.insert(keys, vectors)
        db.build_index(index_type='IVF,Flat', index_params={'nlist': 2})

        dist, idx = db.search(vectors, 2, {'nprobe': 2})
        dist = np.array(dist)
        idx = np.array(idx)
        assert idx.shape == dist.shape
        assert idx.shape == (4, 2)
        np.testing.assert_equal(idx, np.array([[0, 1], [1, 0], [2, 1], [3, 2]]))


@pytest.mark.skip
@setup_tear_down_decorator
def test_milvus_indexer():
    collection_name = 'milvus_indexer'
    create_collection(collection_name)

    vectors = np.array([[1, 1, 1],
                        [10, 10, 10],
                        [100, 100, 100],
                        [1000, 1000, 1000]])
    keys = np.array([0, 1, 2, 3]).reshape(-1, 1)
    with MilvusIndexer(host=host, port=port,
                       collection_name=collection_name, index_type='IVF,Flat',
                       index_params={'nlist': 2}) as indexer:
        indexer.add(keys, vectors)
        idx, dist = indexer.query(vectors, 2, search_params={'nprobe': 2})
        dist = np.array(dist)
        idx = np.array(idx)
        assert idx.shape == dist.shape
        assert idx.shape == (4, 2)
        np.testing.assert_equal(idx, np.array([[0, 1], [1, 0], [2, 1], [3, 2]]))



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from jina.executors.decorators import batching
from jina.executors.encoders import BaseNumericEncoder


class TSNEEncoder(BaseNumericEncoder):
    """
    :class:`TSNEEncoder` encodes data from an ndarray in size `B x T` into an ndarray in size `B x D`.
    https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html
    TSNE does not inherit Transform encoder because it can't have a transform without fit.
    """

    def __init__(self, output_dim: int,
                 random_state: int = 2020,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_dim = output_dim
        self.random_state = random_state

    def post_init(self):
        super().post_init()
        from sklearn.manifold import TSNE
        self.model = TSNE(n_components=self.output_dim, random_state=self.random_state)

    @batching
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        :param data: a `B x T` numpy ``ndarray``, `B` is the size of the batch
        :return: a `B x D` numpy ``ndarray``
        """
        return self.model.fit_transform(data)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Any

from .. import BaseExecutor
from ..compound import CompoundExecutor

if False:
    # fix type-hint complain for sphinx and flake
    import numpy as np


class BaseEncoder(BaseExecutor):
    """``BaseEncoder`` encodes chunk into vector representation.

    The key function is :func:`encode`.

    .. seealso::
        :mod:`jina.drivers.handlers.encode`
    """

    def encode(self, data: Any, *args, **kwargs) -> Any:
        raise NotImplementedError


class BaseNumericEncoder(BaseEncoder):
    """BaseNumericEncoder encodes data from a ndarray, potentially B x ([T] x D) into a ndarray of B x D"""

    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        :param data: a `B x ([T] x D)` numpy ``ndarray``, `B` is the size of the batch
        :return: a `B x D` numpy ``ndarray``
        """
        raise NotImplementedError


class BaseImageEncoder(BaseNumericEncoder):
    """BaseImageEncoder encodes data from a ndarray, potentially B x (Height x Width) into a ndarray of B x D"""
    pass


class BaseVideoEncoder(BaseNumericEncoder):
    """BaseVideoEncoder encodes data from a ndarray, potentially B x (Time x Height x Width) into a ndarray of B x D"""
    pass


class BaseAudioEncoder(BaseNumericEncoder):
    """BaseAudioEncoder encodes data from a ndarray, potentially B x (Time x D) into a ndarray of B x D"""
    pass


class BaseTextEncoder(BaseEncoder):
    """
    BaseTextEncoder encodes data from an array of string type (data.dtype.kind == 'U') of size B into a ndarray of B x D.

    """

    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """

        :param data: an 1d array of string type (data.dtype.kind == 'U') in size B
        :return: an ndarray of `B x D`
        """
        raise NotImplementedError


class PipelineEncoder(CompoundExecutor):
    def encode(self, data: Any, *args, **kwargs) -> Any:
        if not self.components:
            raise NotImplementedError
        for be in self.components:
            data = be.encode(data, *args, **kwargs)
        return data

    def train(self, data, *args, **kwargs):
        if not self.components:
            raise NotImplementedError
        for idx, be in enumerate(self.components):
            if not be.is_trained:
                be.train(data, *args, **kwargs)

            if idx + 1 < len(self.components):
                data = be.encode(data, *args, **kwargs)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple

from .. import QuerySetReader, BaseRecursiveDriver

if False:
    from ...proto import jina_pb2


class ExcludeQL(QuerySetReader, BaseRecursiveDriver):
    """Clean some fields from the document-level protobuf to reduce the total size of the request
        Example::
        - !ExcludeQL
        with:
            fields:
                - chunks
                - buffer

        ExcludeQL will avoid `buffer` and `chunks` fields to be sent to the next `Pod`
    """

    def __init__(self, fields: Tuple, *args, **kwargs):
        """

        :param fields: the pruned field names in tuple
        """
        super().__init__(*args, **kwargs)
        if isinstance(fields, str):
            self._fields = {fields, }
        else:
            self._fields = set(fields)

        # for deleting field in a recursive structure, postorder is safer
        self.recursion_order = 'post'
        self.is_apply_all = False

    def _apply(self, doc: 'jina_pb2.Document', *args, **kwargs):
        for k in self.fields:
            doc.ClearField(k)


class SelectQL(ExcludeQL):
    """Selects some fields from the chunk-level protobuf to reduce the total size of the request, it works with the opposite
    logic as `:class:`ExcludeQL`

        Example::
        - !SelectQL
        with:
            fields:
                - matches

        SelectQL will ensure that the `outgoing` documents only contain the field `matches`
    """
    def _apply(self, doc: 'jina_pb2.Document', *args, **kwargs):
        for k in doc.DESCRIPTOR.fields_by_name.keys():
            if k not in self.fields:
                doc.ClearField(k)


class ExcludeReqQL(ExcludeQL):
    """Clean up request from the request-level protobuf message to reduce the total size of the message

        This is often useful when the proceeding Pods require only a signal, not the full message.
    """

    def __call__(self, *args, **kwargs):
        for k in self.fields:
            self.msg.ClearField(k)


class SelectReqQL(ExcludeReqQL):
    """Clean up request from the request-level protobuf message to reduce the total size of the message, it works with the opposite
    logic as `:class:`ExcludeReqQL`
    """
    def __call__(self, *args, **kwargs):
        for k in self.msg.DESCRIPTOR.fields_by_name.keys():
            if k not in self.fields:
                self.msg.ClearField(k)



from tests import JinaTestCase


class JinaImageTestCase(JinaTestCase):
    @staticmethod
    def create_test_image(output_fn, size_width=50, size_height=50):
        from PIL import Image
        image = Image.new('RGB', size=(size_width, size_height), color=(155, 0, 0))
        with open(output_fn, "wb") as f:
            image.save(f, 'jpeg')

    @staticmethod
    def create_random_img_array(img_height, img_width):
        import numpy as np
        return np.random.randint(0, 256, (img_height, img_width, 3))

    @staticmethod
    def create_test_img_array():
        import numpy as np
        img = np.array([i for i in range(100)]).reshape(10, 10)
        return np.repeat(img[:, :, np.newaxis], 3, axis=2)



import numpy as np

from .. import ArrayStringReader


class TestClass:
    def test_arraystringreader(self):
        """here is my test code

        https://docs.pytest.org/en/stable/getting-started.html#create-your-first-test
        """
        size = 8
        sample_array = np.random.rand(size).astype('float32')
        text = ','.join([str(x) for x in sample_array])

        reader = ArrayStringReader()
        crafted_doc = reader.craft(text, 0)

        assert crafted_doc['blob'].shape[0] == size
        np.testing.assert_array_equal(crafted_doc['blob'], sample_array)



from .. import DeepSegmenter


class TestClass:
    def test_deepsegmenter(self):
        """here is my test code

        https://docs.pytest.org/en/stable/getting-started.html#create-your-first-test
        """



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import glob
import tempfile
import urllib.parse
import webbrowser
from typing import Dict

from .checker import *
from .helper import get_default_login
from ..clients.python import ProgressBar
from ..helper import colored, get_readable_size
from ..logging import get_logger
from ..logging.profile import TimeContext

if False:
    import argparse

_allowed = {'name', 'description', 'author', 'url',
            'documentation', 'version', 'vendor', 'license', 'avatar',
            'platform', 'update', 'keywords'}

_repo_prefix = 'jinahub/'
_label_prefix = 'ai.jina.hub.'


class HubIO:
    """ :class:`HubIO` provides the way to interact with Jina Hub registry.
    You can use it with CLI to package a directory into a Jina Hub image and publish it to the world.

    Examples:
        - :command:`jina hub build my_pod/` build the image
        - :command:`jina hub build my_pod/ --push` build the image and push to the public registry
        - :command:`jina hub pull jinahub/pod.dummy_mwu_encoder:0.0.6` to download the image
    """

    def __init__(self, args: 'argparse.Namespace'):
        self.logger = get_logger(self.__class__.__name__, **vars(args))
        self.args = args
        try:
            import docker
            from docker import APIClient

            self._client = docker.from_env()

            # low-level client
            self._raw_client = APIClient(base_url='unix://var/run/docker.sock')
        except (ImportError, ModuleNotFoundError):
            self.logger.critical('requires "docker" dependency, please install it via "pip install jina[docker]"')
            raise

    def new(self):
        """Create a new executor using cookiecutter template """
        try:
            from cookiecutter.main import cookiecutter
        except (ImportError, ModuleNotFoundError):
            self.logger.critical('requires "cookiecutter" dependency, please install it via "pip install cookiecutter"')
            raise

        cookiecutter(self.args.template, overwrite_if_exists=self.args.overwrite, output_dir=self.args.output_dir)

    def push(self, name: str = None, readme_path: str = None):
        """A wrapper of docker push """
        name = name or self.args.name
        check_registry(self.args.registry, name, _repo_prefix)
        self._check_docker_image(name)
        self.login()
        with ProgressBar(task_name=f'pushing {name}', batch_unit='') as t:
            for line in self._client.images.push(name, stream=True, decode=True):
                t.update(1)
                self.logger.debug(line)
        self.logger.success(f'🎉 {name} is now published!')

        if False and readme_path:
            # unfortunately Docker Hub Personal Access Tokens cannot be used as they are not supported by the API
            _volumes = {os.path.dirname(os.path.abspath(readme_path)): {'bind': '/workspace'}}
            _env = get_default_login()
            _env = {
                'DOCKERHUB_USERNAME': _env['username'],
                'DOCKERHUB_PASSWORD': _env['password'],
                'DOCKERHUB_REPOSITORY': name.split(':')[0],
                'README_FILEPATH': '/workspace/README.md',
            }

            self._client.containers.run('peterevans/dockerhub-description:2.1',
                                        auto_remove=True,
                                        volumes=_volumes,
                                        environment=_env)

        share_link = f'https://api.jina.ai/hub/?jh={urllib.parse.quote_plus(name)}'

        try:
            webbrowser.open(share_link, new=2)
        except:
            pass
        finally:
            self.logger.info(
                f'Check out the usage {colored(share_link, "cyan", attrs=["underline"])} and share it with others!')

    def pull(self):
        """A wrapper of docker pull """
        check_registry(self.args.registry, self.args.name, _repo_prefix)
        self.login()
        with TimeContext(f'pulling {self.args.name}', self.logger):
            image = self._client.images.pull(self.args.name)
        if isinstance(image, list):
            image = image[0]
        image_tag = image.tags[0] if image.tags else ""
        self.logger.success(
            f'🎉 pulled {image_tag} ({image.short_id}) uncompressed size: {get_readable_size(image.attrs["Size"])}')

    def _check_docker_image(self, name: str):
        # check local image
        image = self._client.images.get(name)
        for r in _allowed:
            if f'{_label_prefix}{r}' not in image.labels.keys():
                self.logger.warning(f'{r} is missing in your docker image labels, you may want to check it')
        try:
            if name != safe_url_name(
                    f'{_repo_prefix}' + '{type}.{kind}.{name}:{version}'.format(
                        **{k.replace(_label_prefix, ''): v for k, v in image.labels.items()})):
                raise ValueError(f'image {name} does not match with label info in the image')
        except KeyError:
            self.logger.error('missing key in the label of the image')
            raise

        self.logger.info(f'✅ {name} is a valid Jina Hub image, ready to publish')

    def login(self):
        """A wrapper of docker login """
        try:
            password = self.args.password  # or (self.args.password_stdin and self.args.password_stdin.read())
        except ValueError:
            password = ''

        if self.args.username and password:
            self._client.login(username=self.args.username, password=password,
                               registry=self.args.registry)
        else:
            # use default login
            self._client.login(**get_default_login(), registry=self.args.registry)

    def build(self):
        """A wrapper of docker build """
        self._check_completeness()
        is_build_success = True
        with TimeContext(f'building {colored(self.canonical_name, "green")}', self.logger):

            streamer = self._raw_client.build(
                decode=True,
                path=self.args.path,
                tag=self.canonical_name,
                pull=self.args.pull,
                dockerfile=self.dockerfile_path_revised,
                rm=True
            )

            for chunk in streamer:
                if 'stream' in chunk:
                    for line in chunk['stream'].splitlines():
                        if 'error' in line.lower():
                            self.logger.critical(line)
                            is_build_success = False
                        elif 'warning' in line.lower():
                            self.logger.warning(line)
                        else:
                            self.logger.info(line)

        if is_build_success:
            # compile it again, but this time don't show the log
            image, log = self._client.images.build(path=self.args.path,
                                                   tag=self.canonical_name,
                                                   pull=self.args.pull,
                                                   dockerfile=self.dockerfile_path_revised,
                                                   rm=True)

            # success
            self.logger.success(
                f'🎉 built {image.tags[0]} ({image.short_id}) uncompressed size: {get_readable_size(image.attrs["Size"])}')

            if self.args.push:
                self.push(image.tags[0], self.readme_path)
        else:
            self.logger.error(f'can not build the image, please double check the log')

    def _check_completeness(self):
        self.dockerfile_path = get_exist_path(self.args.path, 'Dockerfile')
        self.manifest_path = get_exist_path(self.args.path, 'manifest.yml')
        self.readme_path = get_exist_path(self.args.path, 'README.md')
        self.requirements_path = get_exist_path(self.args.path, 'requirements.txt')

        yaml_glob = glob.glob(os.path.join(self.args.path, '*.yml'))
        if yaml_glob:
            yaml_glob.remove(self.manifest_path)

        py_glob = glob.glob(os.path.join(self.args.path, '*.py'))

        test_glob = glob.glob(os.path.join(self.args.path, 'tests/test_*.py'))

        completeness = {
            'Dockerfile': self.dockerfile_path,
            'manifest.yml': self.manifest_path,
            'README.md': os.path.exists(self.readme_path),
            'requirements.txt': os.path.exists(self.readme_path),
            '*.yml': yaml_glob,
            '*.py': py_glob,
            'tests': test_glob
        }

        self.logger.info(
            f'a completeness check\n' +
            '\n'.join('%4s %-20s %s' % (colored('✓', 'green') if v else colored('✗', 'red'), k, v) for k, v in
                      completeness.items()) + '\n')

        if completeness['Dockerfile'] and completeness['manifest.yml']:
            pass
        else:
            self.logger.critical('Dockerfile or manifest.yml is not given, can not build')
            raise FileNotFoundError('Dockerfile or manifest.yml is not given, can not build')

        tmp = self._read_manifest(self.manifest_path)
        self.dockerfile_path_revised = self._get_revised_dockerfile(self.dockerfile_path, tmp)
        self.canonical_name = safe_url_name(f'{_repo_prefix}' + '{type}.{kind}.{name}:{version}'.format(**tmp))

    def _read_manifest(self, path: str, validate: bool = True):
        with resource_stream('jina', '/'.join(('resources', 'hub-builder', 'manifest.yml'))) as fp:
            tmp = yaml.load(fp)  # do not expand variables at here, i.e. DO NOT USE expand_dict(yaml.load(fp))

        with open(path) as fp:
            tmp.update(yaml.load(fp))

        if validate:
            self._validate_manifest(tmp)

        return tmp

    def _validate_manifest(self, manifest: Dict):
        required = {'name', 'type', 'version'}

        # check the required field in manifest
        for r in required:
            if r not in manifest:
                raise ValueError(f'{r} is missing in the manifest.yaml, it is required')

        # check if all fields are there
        for r in _allowed:
            if r not in manifest:
                self.logger.warning(f'{r} is missing in your manifest.yml, you may want to check it')

        # check name
        check_name(manifest['name'])
        # check_image_type
        check_image_type(manifest['type'])
        # check version number
        check_version(manifest['version'])
        # check version number
        check_license(manifest['license'])
        # check platform
        if not isinstance(manifest['platform'], list):
            manifest['platform'] = list(manifest['platform'])
        check_platform(manifest['platform'])

        # replace all chars in value to safe chars
        for k, v in manifest.items():
            if v and isinstance(v, str):
                manifest[k] = remove_control_characters(v)
            elif v and isinstance(v, list):
                manifest[k] = ','.join(v)

        # show manifest key-values
        for k, v in manifest.items():
            self.logger.debug(f'{k}: {v}')

    def _get_revised_dockerfile(self, dockerfile_path: str, manifest: Dict):
        # modify dockerfile
        revised_dockerfile = []
        with open(dockerfile_path) as fp:
            for l in fp:
                revised_dockerfile.append(l)
                if l.startswith('FROM'):
                    revised_dockerfile.append('LABEL ')
                    revised_dockerfile.append(
                        ' \\      \n'.join(f'{_label_prefix}{k}="{v}"' for k, v in manifest.items()))

        f = tempfile.NamedTemporaryFile('w', delete=False).name
        with open(f, 'w', encoding='utf8') as fp:
            fp.writelines(revised_dockerfile)

        for k in revised_dockerfile:
            self.logger.debug(k)
        return f



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterable

from . import BaseExecutableDriver, QuerySetReader
from .helper import extract_docs

if False:
    from ..proto import jina_pb2


class BaseSearchDriver(BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`craft` by default """

    def __init__(self, executor: str = None, method: str = 'query', *args, **kwargs):
        super().__init__(executor, method, *args, **kwargs)
        self._is_apply = False


class KVSearchDriver(BaseSearchDriver):
    """Fill in the doc/chunk-level top-k results using the :class:`jina.executors.indexers.meta.BasePbIndexer`

    .. warning::
        This driver loops over all chunk/chunk's top-K results, each step fires a query.
        This may not be very efficient, as the total number of queries depends on ``level``

             - ``level=chunk``: D x C x K
             - ``level=doc``: D x K
             - ``level=all``: D x C x K

        where:
            - D is the number of queries
            - C is the number of chunks per query/doc
            - K is the top-k
    """

    def _apply_all(self, docs: Iterable['jina_pb2.Document'], *args, **kwargs):
        miss_idx = []  #: missed hit results, some search may not end with results. especially in shards
        for idx, retrieved_doc in enumerate(docs):
            r = self.exec_fn(retrieved_doc.id)
            if r:
                retrieved_doc.MergeFrom(r)
            else:
                miss_idx.append(idx)

        # delete non-existed matches in reverse
        for j in reversed(miss_idx):
            del docs[j]


class VectorSearchDriver(QuerySetReader, BaseSearchDriver):
    """Extract chunk-level embeddings from the request and use the executor to query it

    """

    def __init__(self, top_k: int = 50, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._top_k = top_k

    def _apply_all(self, docs: Iterable['jina_pb2.Document'], *args, **kwargs):
        embed_vecs, doc_pts, bad_doc_ids = extract_docs(docs, embedding=True)

        if bad_doc_ids:
            self.logger.warning(f'these bad docs can not be added: {bad_doc_ids}')

        if doc_pts:
            idx, dist = self.exec_fn(embed_vecs, top_k=self.top_k)
            op_name = self.exec.__class__.__name__
            for doc, topks, scores in zip(doc_pts, idx, dist):
                for match_id, score in zip(topks, scores):
                    r = doc.matches.add()
                    r.level_depth = doc.level_depth
                    r.id = match_id
                    r.score.ref_id = doc.id
                    r.score.value = score
                    r.score.op_name = op_name



import numpy as np

from .. import AudioMonophoner


class TestClass:
    def test_audiomonophoner(self):
        """here is my test code

        https://docs.pytest.org/en/stable/getting-started.html#create-your-first-test
        """
        signal_orig = np.random.randn(2, 31337)

        crafter = AudioMonophoner()
        crafted_doc = crafter.craft(signal_orig, 0)

        signal_mono = crafted_doc["blob"]
        assert signal_mono.shape[0] == signal_orig.shape[1]



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import time
from typing import Iterator, Callable, Union

from . import request
from .grpc import GrpcClient
from .helper import ProgressBar
from ...enums import ClientMode
from ...excepts import BadClient
from ...logging import default_logger
from ...logging.profile import TimeContext
from ...proto import jina_pb2

if False:
    # fix type-hint complain for sphinx and flake
    import argparse


class PyClient(GrpcClient):
    """A simple Python client for connecting to the gateway. This class is for internal only,
    use the python interface :func:`jina.clients.py_client` to start :class:`PyClient` if you
    want to use it in Python.

    Assuming a Flow is "standby" on 192.168.1.100, with port_expose at 55555.

    .. highlight:: python
    .. code-block:: python

        from jina.clients import py_client

        # to test connectivity
        py_client(host='192.168.1.100', port_expose=55555).dry_run()

        # to search
        py_client(host='192.168.1.100', port_expose=55555).search(input_fn, output_fn)

        # to index
        py_client(host='192.168.1.100', port_expose=55555).index(input_fn, output_fn)

    """

    def __init__(self, args: 'argparse.Namespace'):
        """

        :param args: args provided by the CLI

        """
        super().__init__(args)
        self._mode = self.args.mode
        self._input_fn = None

    @property
    def mode(self) -> str:
        return self._mode

    @mode.setter
    def mode(self, value: ClientMode):
        if isinstance(value, ClientMode):
            self._mode = value
            self.args.mode = value
        else:
            raise ValueError(f'{value} must be one of {ClientMode}')

    @staticmethod
    def check_input(input_fn: Union[Iterator['jina_pb2.Document'], Iterator[bytes], Callable] = None):
        """Validate the input_fn and print the first request if success

        :param input_fn: the input function
        """
        if hasattr(input_fn, '__call__'):
            input_fn = input_fn()

        kwargs = {'data': input_fn}

        try:
            r = next(getattr(request, 'index')(**kwargs))
            if r is not None:
                default_logger.success(f'input_fn is valid')
            else:
                raise TypeError
        except:
            default_logger.error(f'input_fn is not valid!')
            raise

    def call_unary(self, data: Union['jina_pb2.Document', bytes], mode: ClientMode) -> None:
        """ Calling the server with one request only, and return the result

        This function should not be used in production due to its low-efficiency. For example,
        you should not use it in a for-loop. Use :meth:`call` instead.
        Nonetheless, you can use it for testing one query and check the result.

        :param data: the binary data of the document or the ``Document`` in protobuf
        :param mode: request will be sent in this mode, available ``train``, ``index``, ``query``
        """
        self.mode = mode
        kwargs = vars(self.args)
        kwargs['data'] = [data]

        req_iter = getattr(request, str(self.mode).lower())(**kwargs)
        return self._stub.CallUnary(next(req_iter))

    def call(self, callback: Callable[['jina_pb2.Message'], None] = None, **kwargs) -> None:
        """ Calling the server, better use :func:`start` instead.

        :param callback: a callback function, invoke after every response is received
        """
        # take the default args from client
        _kwargs = vars(self.args)
        _kwargs['data'] = self.input_fn
        # override by the caller-specific kwargs
        _kwargs.update(kwargs)

        tname = str(self.mode).lower()
        if 'mode' in kwargs:
            tname = str(kwargs['mode']).lower()

        req_iter = getattr(request, tname)(**_kwargs)
        # next(req_iter)

        with ProgressBar(task_name=tname) as p_bar, TimeContext(tname):
            for resp in self._stub.Call(req_iter):
                if resp.status.code >= jina_pb2.Status.ERROR:
                    self.logger.error(f'callback() may not work properly '
                                      f'due to the bad response: {resp.status.description}')
                    self.logger.error(resp.status.details)
                if callback:
                    try:
                        if self.args.callback_on_body:
                            resp = getattr(resp, resp.WhichOneof('body'))
                        callback(resp)
                    except Exception as ex:
                        raise BadClient('error in client\'s callback: %s' % ex)
                p_bar.update(self.args.batch_size)

    @property
    def input_fn(self) -> Union[Iterator['jina_pb2.Document'], Iterator[bytes], Callable]:
        """ An iterator of bytes, each element represents a document's raw content,
        i.e. ``input_fn`` defined in the protobuf
        """
        if self._input_fn is not None:
            return self._input_fn
        else:
            raise BadClient('input_fn is empty or not set')

    @input_fn.setter
    def input_fn(self, bytes_gen: Union[Iterator['jina_pb2.Document'], Iterator[bytes], Callable]):
        if self._input_fn:
            self.logger.warning('input_fn is not empty, overrided')
        if hasattr(bytes_gen, '__call__'):
            self._input_fn = bytes_gen()
        else:
            self._input_fn = bytes_gen

    def dry_run(self, as_request: str) -> bool:
        """A dry run request is a Search/Index/Train Request with empty content.
        Useful for testing connectivity and debugging the connectivity of the server/flow

        :param as_request: send the dry run request as one of 'index', 'search', 'train', 'control' request
        :return: if dry run is successful or not
        """

        def req_gen():
            req = jina_pb2.Request()
            if as_request == 'train':
                req.train.CopyFrom(jina_pb2.Request.TrainRequest())
            elif as_request == 'index':
                req.index.CopyFrom(jina_pb2.Request.IndexRequest())
            elif as_request == 'search':
                req.search.CopyFrom(jina_pb2.Request.SearchRequest())
            elif as_request == 'control':
                req.control.CopyFrom(jina_pb2.Request.ControlRequest())
            else:
                raise ValueError(
                    f'as_request={as_request} is not supported, must be one of "train", "search", "index", "control"')
            yield req

        before = time.perf_counter()
        for resp in self._stub.Call(req_gen()):
            if resp.status.code < jina_pb2.Status.ERROR:
                self.logger.info(
                    f'dry run of {as_request} takes {time.perf_counter() - before:.3f}s, this flow has a good connectivity')
                return True
            else:
                self.logger.error(resp.status)

        return False

    def train(self, input_fn: Union[Iterator['jina_pb2.Document'], Iterator[bytes], Callable] = None,
              output_fn: Callable[['jina_pb2.Message'], None] = None, **kwargs):
        self.mode = ClientMode.TRAIN
        self.input_fn = input_fn
        if not self.args.skip_dry_run:
            self.dry_run(as_request='train')
        self.start(output_fn, **kwargs)

    def search(self, input_fn: Union[Iterator['jina_pb2.Document'], Iterator[bytes], Callable] = None,
               output_fn: Callable[['jina_pb2.Message'], None] = None, **kwargs):
        self.mode = ClientMode.SEARCH
        self.input_fn = input_fn
        if not self.args.skip_dry_run:
            self.dry_run(as_request='search')
        self.start(output_fn, **kwargs)

    def index(self, input_fn: Union[Iterator['jina_pb2.Document'], Iterator[bytes], Callable] = None,
              output_fn: Callable[['jina_pb2.Message'], None] = None, **kwargs):
        self.mode = ClientMode.INDEX
        self.input_fn = input_fn
        if not self.args.skip_dry_run:
            self.dry_run(as_request='index')
        self.start(output_fn, **kwargs)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os

from .base import get_logger

default_logger = get_logger('JINA')  #: a logger at the global-level

if 'JINA_LOG_PROFILING' in os.environ:
    profile_logger = get_logger('PROFILE', log_profile=True)  #: a logger for profiling
    default_logger.success('profiling is enabled')
else:
    profile_logger = default_logger



from . import JinaImageTestCase
from .. import SlidingWindowImageCropper


class ImageSegmentTestCase(JinaImageTestCase):

    def test_sliding_windows_no_padding(self):
        img_size = 14
        img_array = self.create_random_img_array(img_size, img_size)
        output_dim = 4
        strides = (6, 6)
        crafter = SlidingWindowImageCropper(target_size=output_dim, strides=strides, padding=False)
        chunks = crafter.craft(img_array)
        assert len(chunks) == 4
        assert chunks[0]['location'] == (0, 0)
        assert chunks[1]['location'] == (0, 6)
        assert chunks[2]['location'] == (6, 0)
        assert chunks[3]['location'] == (6, 6)

    def test_sliding_windows_with_padding(self):
        img_size = 14
        img_array = self.create_random_img_array(img_size, img_size)
        output_dim = 4
        strides = (6, 6)
        crafter = SlidingWindowImageCropper(target_size=output_dim, strides=strides, padding=True)
        chunks = crafter.craft(img_array)
        assert len(chunks) == 9
        assert chunks[0]['location'] == (0, 0)
        assert chunks[1]['location'] == (0, 6)
        assert chunks[2]['location'] == (0, 12)
        assert chunks[3]['location'] == (6, 0)
        assert chunks[4]['location'] == (6, 6)
        assert chunks[5]['location'] == (6, 12)
        assert chunks[6]['location'] == (12, 0)
        assert chunks[7]['location'] == (12, 6)
        assert chunks[8]['location'] == (12, 12)

    def test_sliding_windows_without_padding_rectangular_ugly_shapes(self):
        height = 16
        width = 11
        img_array = self.create_random_img_array(img_height=height, img_width=width)
        output_dim = 4
        strides = (4, 4)
        crafter = SlidingWindowImageCropper(target_size=output_dim, strides=strides, padding=False)
        chunks = crafter.craft(img_array)
        assert len(chunks) == 8
        assert chunks[0]['location'] == (0, 0)
        assert chunks[1]['location'] == (0, 4)
        assert chunks[2]['location'] == (4, 0)
        assert chunks[3]['location'] == (4, 4)
        assert chunks[4]['location'] == (8, 0)
        assert chunks[5]['location'] == (8, 4)
        assert chunks[6]['location'] == (12, 0)
        assert chunks[7]['location'] == (12, 4)
